from __future__ import annotations

import torch
import torch.nn as nn
import triton
import triton.language as tl

from pyjuice.layer import SumLayer, ProdLayer, InputLayer
from pyjuice.utils.kernel_launcher import triton_jit


@triton_jit
def prod_layer_partition_fn_kernel(node_mars, element_mars, nids, cids, n_num_nblocks: tl.constexpr, n_num_nodes: tl.constexpr,
                                   c_num_nblocks: tl.constexpr, block_size: tl.constexpr, TILE_SIZE_N: tl.constexpr, TILE_SIZE_C: tl.constexpr):
    pid_n = tl.program_id(0)

    n_ids = tl.arange(0, TILE_SIZE_N) + pid_n * TILE_SIZE_N
    n_mask = (n_ids < n_num_nodes)
    group_ids = n_ids // block_size
    offs_node = n_ids % block_size

    c_ids = tl.arange(0, TILE_SIZE_C)
    c_mask = (c_ids < c_num_nblocks)

    nc_mask = n_mask[:,None] & c_mask[None,:]

    cids_base = tl.load(cids + group_ids[:,None] * c_num_nblocks + c_ids[None,:], mask = nc_mask)
    cids = cids_base + offs_node[:,None]
    cmars = tl.load(node_mars + cids, mask = nc_mask, other = 0.0)

    nmars = tl.sum(cmars, axis = 1)

    nids_base = tl.load(nids + group_ids, mask = n_mask)
    nids = nids_base + offs_node
    tl.store(element_mars + nids, nmars, mask = n_mask)


def prod_layer_partition_fn(layer, node_mars, element_mars, params):
    block_size = layer.block_size
    for partition_id in range(layer.num_fw_partitions):
        nids = layer.partitioned_nids[partition_id]
        cids = layer.partitioned_cids[partition_id]

        n_num_nblocks = nids.size(0)
        n_num_nodes = n_num_nblocks * block_size
        c_num_nblocks = cids.size(1)

        if c_num_nblocks <= 2048:
            TILE_SIZE_C = triton.next_power_of_2(c_num_nblocks)
            TILE_SIZE_N = min(2048 // TILE_SIZE_C, triton.next_power_of_2(n_num_nodes))

            grid = (triton.cdiv(n_num_nodes, TILE_SIZE_N),)

            prod_layer_partition_fn_kernel[grid](
                node_mars,
                element_mars,
                nids,
                cids,
                n_num_nblocks = n_num_nblocks,
                n_num_nodes = n_num_nodes,
                c_num_nblocks = c_num_nblocks,
                block_size = block_size,
                TILE_SIZE_N = TILE_SIZE_N,
                TILE_SIZE_C = TILE_SIZE_C
            )

        else:
            raise NotImplementedError("Need to implement a new kernel for this case.")


@triton_jit
def sum_layer_partition_fn_block_kernel(node_mars, element_mars, mparams, nids, cids, pids, n_num_nodes: tl.constexpr,
                                        c_num_nodes: tl.constexpr, block_size: tl.constexpr, TILE_SIZE_N: tl.constexpr, 
                                        TILE_SIZE_C: tl.constexpr, C_NUM_TILES: tl.constexpr):
    pid_n = tl.program_id(0)

    ngroup_id = (pid_n * TILE_SIZE_N) // block_size
    offs_nnode = tl.arange(0, TILE_SIZE_N) + ((pid_n * TILE_SIZE_N) % block_size)

    acc = tl.zeros([TILE_SIZE_N], dtype = tl.float32) - float("inf")
    for pid_c in range(C_NUM_TILES):

        offs_cnode = tl.arange(0, TILE_SIZE_C) + pid_c * TILE_SIZE_C
        mask_cnode = (offs_cnode < c_num_nodes)
        cnode_ids = tl.load(cids + ngroup_id * c_num_nodes + offs_cnode, mask = mask_cnode) # [TILE_SIZE_C]
        cmars = tl.load(element_mars + cnode_ids, mask = mask_cnode, other = 0.0)

        par_ids = tl.load(pids + ngroup_id * c_num_nodes + offs_cnode, mask = mask_cnode) # [TILE_SIZE_C]
        pars = tl.load(mparams + par_ids[None,:] + offs_nnode[:,None], mask = mask_cnode[None,:]) # [TILE_SIZE_N, TILE_SIZE_C]

        cmars_max = tl.max(cmars)
        cmars_sub = tl.exp(cmars - cmars_max)

        nmars = tl.sum(pars * cmars_sub[None,:], axis = 1)

        acc = tl.where(cmars_max > acc,
            tl.log(nmars + tl.exp(acc - cmars_max) + 1e-24) + cmars_max,
            tl.log(tl.exp(cmars_max - acc) * nmars + 1.0) + acc
        )

    nid_base = tl.load(nids + ngroup_id)
    nnode_ids = nid_base + offs_nnode # [TILE_SIZE_N]
    tl.store(node_mars + nnode_ids, acc)


@triton_jit
def sum_layer_partition_fn_node_kernel(node_mars, element_mars, mparams, nids, cids, pids, n_num_nodes: tl.constexpr,
                                       c_num_nodes: tl.constexpr, block_size: tl.constexpr, TILE_SIZE_N: tl.constexpr, 
                                       TILE_SIZE_C: tl.constexpr, C_NUM_TILES: tl.constexpr):
    pid_n = tl.program_id(0)

    n_ids = tl.arange(0, TILE_SIZE_N) + pid_n * TILE_SIZE_N
    n_mask = (n_ids < n_num_nodes)
    ngroup_ids = n_ids // block_size
    offs_nnode = n_ids % block_size

    acc = tl.zeros([TILE_SIZE_N], dtype = tl.float32) - float("inf")
    for pid_c in range(C_NUM_TILES):

        offs_cnode = tl.arange(0, TILE_SIZE_C) + pid_c * TILE_SIZE_C
        mask_cnode = (offs_cnode < c_num_nodes)
        cnode_ids = tl.load(cids + ngroup_ids[:,None] * c_num_nodes + offs_cnode[None,:], mask = (n_mask[:,None] & mask_cnode[None,:])) # [TILE_SIZE_N, TILE_SIZE_C]
        cmars = tl.load(element_mars + cnode_ids, mask = (n_mask[:,None] & mask_cnode[None,:]), other = 0.0)

        par_ids = tl.load(pids + ngroup_ids[:,None] * c_num_nodes + offs_cnode[None,:], mask = (n_mask[:,None] & mask_cnode[None,:])) # [TILE_SIZE_N, TILE_SIZE_C]
        pars = tl.load(mparams + par_ids + offs_nnode[:,None], mask = (n_mask[:,None] & mask_cnode[None,:])) # [TILE_SIZE_N, TILE_SIZE_C]

        cmars_max = tl.max(cmars, axis = 1)
        cmars_sub = tl.exp(cmars - cmars_max[:,None])

        nmars = tl.sum(pars * cmars_sub, axis = 1)

        acc = tl.where(cmars_max > acc,
            tl.log(nmars + tl.exp(acc - cmars_max) + 1e-24) + cmars_max,
            tl.log(tl.exp(cmars_max - acc) * nmars + 1.0) + acc
        )

    nid_base = tl.load(nids + ngroup_ids, mask = n_mask) # [TILE_SIZE_N]
    nnode_ids = nid_base + offs_nnode # [TILE_SIZE_N]
    tl.store(node_mars + nnode_ids, acc, mask = n_mask)


def sum_layer_partition_fn(layer, node_mars, element_mars, params):
    block_size = layer.block_size
    for partition_id in range(layer.num_fw_partitions):
        nids = layer.partitioned_nids[partition_id]
        cids = layer.partitioned_cids[partition_id]
        pids = layer.partitioned_pids[partition_id]

        n_num_nodes = nids.size(0) * block_size
        c_num_nodes = cids.size(1)

        if block_size >= 8:
            TILE_SIZE_C = min(1024, triton.next_power_of_2(c_num_nodes))
            TILE_SIZE_N = min(1024 // TILE_SIZE_C, block_size)

            C_NUM_TILES = triton.cdiv(c_num_nodes, TILE_SIZE_C)
            
            grid = (triton.cdiv(n_num_nodes, TILE_SIZE_N),)

            sum_layer_partition_fn_block_kernel[grid](
                node_mars,
                element_mars,
                params,
                nids,
                cids,
                pids,
                n_num_nodes = n_num_nodes,
                c_num_nodes = c_num_nodes,
                block_size = block_size,
                TILE_SIZE_N = TILE_SIZE_N,
                TILE_SIZE_C = TILE_SIZE_C,
                C_NUM_TILES = C_NUM_TILES
            )

        else:
            TILE_SIZE_C = min(1024, triton.next_power_of_2(c_num_nodes))
            TILE_SIZE_N = min(1024 // TILE_SIZE_C, n_num_nodes)

            C_NUM_TILES = triton.cdiv(c_num_nodes, TILE_SIZE_C)

            grid = (triton.cdiv(n_num_nodes, TILE_SIZE_N),)

            sum_layer_partition_fn_node_kernel[grid](
                node_mars,
                element_mars,
                params,
                nids,
                cids,
                pids,
                n_num_nodes = n_num_nodes,
                c_num_nodes = c_num_nodes,
                block_size = block_size,
                TILE_SIZE_N = TILE_SIZE_N,
                TILE_SIZE_C = TILE_SIZE_C,
                C_NUM_TILES = C_NUM_TILES
            )


def eval_partition_fn(pc):
    """
    Computes the partition function of every node.
    We use the first # nodes entries in `pc.node_mars` and the first # element entries in `pc.element_mars`.
    """
    assert hasattr(pc, "node_mars") and hasattr(pc, "element_mars")
    assert pc.node_mars.numel() >= pc.num_nodes
    assert pc.element_mars.numel() >= pc.num_elements

    node_mars = pc.node_mars.view(-1) # Reuse the allocated memory
    element_mars = pc.element_mars.view(-1)

    node_mars[:pc.num_nodes] = 0.0
    element_mars[:pc.num_elements] = -float("inf")

    # Forward pass over the input layers
    for layer in pc.input_layer_group:
        layer.eval_partition_fn(node_mars)

    # Forward pass over the inner layers
    for layer_id, layer_group in enumerate(pc.inner_layer_groups):
        layer_group = pc.inner_layer_groups[layer_id]

        for layer in layer_group:
            if layer.is_prod():
                prod_layer_partition_fn(layer, node_mars, element_mars, pc.params)

            elif layer.is_sum():
                sum_layer_partition_fn(layer, node_mars, element_mars, pc.params)

    partition_fn = node_mars[pc._root_node_range[0]:pc._root_node_range[1]]
    return partition_fn


from .nodes import CircuitNodes
from .input_nodes import InputNodes
from .prod_nodes import ProdNodes
from .sum_nodes import SumNodes
from .construction import multiply, summate, inputs, set_block_size, structural_properties
from .methods.traversal import foreach, foldup_aggregate


from __future__ import annotations

import numpy as np
import torch
from typing import Sequence, Union, Type, Optional, Dict
from copy import deepcopy

from pyjuice.graph import InputRegionNode
from .distributions import Distribution
from .nodes import CircuitNodes


class InputNodes(CircuitNodes):
    """
    A class representing vectors of input nodes. It is created by :code:`pyjuice.inputs`.

    :param num_node_blocks: number of node blocks
    :type num_node_blocks: int

    :param scope: variable scope (set of variables)
    :type scope: Union[Sequence,BitSet]

    :param dist: input distribution
    :type dist: Distribution

    :param params: parameters of the vector of nodes
    :type params: Optional[Tensor]

    :param block_size: block size
    :type block_size: int
    """
    def __init__(self, num_node_blocks: int, scope: Union[Sequence,BitSet], dist: Distribution, 
                 params: Optional[torch.Tensor] = None, block_size: int = 0, 
                 _no_set_meta_params: bool = False, **kwargs) -> None:

        rg_node = InputRegionNode(scope)
        super(InputNodes, self).__init__(num_node_blocks, rg_node, block_size = block_size, **kwargs)

        self.chs = [] # InputNodes has no children

        self.dist = dist

        # Init parameters and meta-parameters
        if not _no_set_meta_params and self.dist.need_meta_parameters:
            self.set_meta_params(**kwargs)
        if params is not None:
            self.set_params(params)

        # Callbacks
        self._run_init_callbacks(**kwargs)

        # Parameter initialization flag
        self._param_initialized = False

    @property
    def num_edges(self):
        return 0

    def duplicate(self, scope: Optional[Union[int,Sequence,BitSet]] = None, tie_params: bool = False) -> InputNodes:
        """
        Create a duplication of the current node with the same specification (i.e., number of nodes, block size, distribution).

        :param scope: variable scope of the duplication
        :type scope: Optional[Union[int,Sequence,BitSet]]

        :param tie_params: whether to tie the parameters of the current node and the duplicated node
        :type tie_params: bool

        :returns: a duplicated :code:`InputNodes`
        """
        if scope is None:
            scope = self.scope
        else:
            if isinstance(scope, int):
                scope = [scope]

            assert len(scope) == len(self.scope)

        dist = deepcopy(self.dist)

        ns = InputNodes(self.num_node_blocks, scope = scope, dist = dist, block_size = self.block_size, source_node = self if tie_params else None)

        if hasattr(self, "_params") and self._params is not None and not tie_params:
            ns._params = self._params.clone()

        return ns

    def get_params(self) -> torch.Tensor:
        """
        Get the input node parameters.
        """
        if not self.provided("_params"):
            return None
        else:
            return self._params

    def set_params(self, params: Union[torch.Tensor,Dict], normalize: bool = True):
        """
        Set the input node parameters.

        :param params: parameters to be set
        :type params: Union[torch.Tensor,Dict]

        :param normalize: whether to normalize the parameters
        :type normalize: bool
        """
        assert params.numel() == self.num_nodes * self.dist.num_parameters()

        params = params.reshape(-1)
        if normalize:
            params = self.dist.normalize_parameters(params)

        self._param_initialized = True
        self._params = params

    def set_meta_params(self, **kwargs):
        """
        Set the meta-parameters such as the mask of input nodes with the `MaskedCategorical` distribution.
        """
        params = self.dist.set_meta_parameters(self.num_nodes, **kwargs)

        self._param_initialized = False
        self._params = params

    def init_parameters(self, perturbation: float = 2.0, recursive: bool = True, 
                        is_root: bool = True, ret_params: bool = False, **kwargs) -> None:
        """
        Randomly initialize node parameters.

        :param perturbation: "amount of perturbation" added to the parameters (should be greater than 0)
        :type perturbation: float

        :param recursive: whether to recursively apply the function to child nodes
        :type recursive: bool
        """
        if not self.is_tied() and not self.has_params():
            self._params = self.dist.init_parameters(
                num_nodes = self.num_nodes,
                perturbation = perturbation,
                params = self.get_params(),
                **kwargs
            )

            if ret_params:
                return self._params

        elif self.is_tied() and ret_params:
            return self.get_source_ns().init_parameters(
                perturbation = perturbation,
                recursive = False,
                is_root = True,
                ret_params = True,
                **kwargs
            )

    def get_data_dtype(self):
        return self.dist.get_data_dtype()

    def __repr__(self):
        return f"InputNodes(num_node_blocks={self.num_node_blocks}, block_size={self.block_size}, dist={type(self.dist)})"



from __future__ import annotations

import torch
import triton
import triton.language as tl
import math
from typing import Optional, Any

from .distributions import Distribution


class Gaussian(Distribution):
    """
    A class representing Gaussian distributions.

    :note: `mu` and `sigma` are used to specify (approximately) the mean and std of the data. This is used for parameter initialization.

    :note: The parameters will NOT be initialized directly using the values of `mu` and `sigma`, perturbations will be added. You can specify the initialization behavior by passing `perturbation`, `mu`, and `sigma` to the `init_parameters` function.

    :param mu: mean of the Gaussian
    :type mu: float

    :param sigma: standard deviation of the Gaussian
    :type sigma: float
    """

    def __init__(self, mu: Optional[float] = None, sigma: Optional[float] = None, min_sigma: float = 0.01):
        super(Gaussian, self).__init__()

        self.mu = mu
        self.sigma = sigma
        self.min_sigma = min_sigma

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        return "Gaussian"

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return [self.min_sigma]

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        return 2

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        return 3

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[Any] = None, **kwargs):
        """
        Initialize parameters for `num_nodes` nodes.
        Returned parameters should be flattened into a vector.
        """
        if params is not None:
            assert isinstance(params, torch.Tensor)
            assert params.numel() == num_nodes * self.num_parameters()
            return params.contiguous()

        mu = kwargs["mu"] if "mu" in kwargs else self.mu
        sigma = kwargs["sigma"] if "sigma" in kwargs else self.sigma
        assert (mu is not None), "`mu` should be provided either during initialization or when calling `init_parameters`."
        assert (sigma is not None), "`sigma` should be provided either during initialization or when calling `init_parameters`."

        mus = torch.normal(mean = torch.ones([num_nodes]) * mu, std = sigma)
        sigmas = (torch.exp(torch.rand([num_nodes]) * -0.04 * perturbation) * sigma).clamp(min = self.min_sigma)

        return torch.stack((mus, sigmas), dim = 1).reshape(-1).contiguous()

    @staticmethod
    def fw_mar_fn(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        sigma = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0)

        centered_val = (data - mu) / sigma
        log_probs = -0.5 * centered_val * centered_val - 0.5 * 1.8378770664093453 - tl.log(sigma) # the constant is `math.log(2 * math.pi)`

        return log_probs

    @staticmethod
    def bk_flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                   s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE):
        stat1 = data * flows
        stat2 = data * data * flows
        stat3 = flows

        tl.atomic_add(param_flows_ptr + s_pfids, stat1, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 1, stat2, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 2, stat3, mask = mask)

    @staticmethod
    def bk_flow_mask_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                        s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE, TILE_SIZE_K):
        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        sigma = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0)

        stat1 = mu * flows
        stat2 = (sigma * sigma + mu * mu) * flows
        stat3 = flows

        tl.atomic_add(param_flows_ptr + s_pfids, stat1, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 1, stat2, mask = mask)
        tl.atomic_add(param_flows_ptr + s_pfids + 2, stat3, mask = mask)

    @staticmethod
    def sample_fn(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed):

        rnd_val = tl.randn(seed, tl.arange(0, BLOCK_SIZE))

        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        sigma = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0)
        
        sample = mu + rnd_val * sigma

        # Write back to `samples`
        sample_offsets = vids * batch_size + batch_offsets
        tl.store(samples_ptr + sample_offsets, sample, mask = mask)

    @staticmethod
    def em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE):
        # Get `min_sigma` from `metadata`
        s_mids = tl.load(s_mids_ptr + local_offsets, mask = mask, other = 0)
        min_sigma = tl.load(metadata_ptr + s_mids, mask = mask, other = 0)

        mu = tl.load(params_ptr + s_pids, mask = mask, other = 0)
        sigma = tl.load(params_ptr + s_pids + 1, mask = mask, other = 0)
        ori_theta1 = mu
        ori_theta2 = sigma * sigma + mu * mu

        stat1 = tl.load(param_flows_ptr + s_pfids, mask = mask)
        stat2 = tl.load(param_flows_ptr + s_pfids + 1, mask = mask)
        stat3 = tl.load(param_flows_ptr + s_pfids + 2, mask = mask)

        new_theta1 = stat1 / (stat3 + 1e-10)
        new_theta2 = stat2 / (stat3 + 1e-10)

        # Get the updated natural parameters
        updated_theta1 = (1.0 - step_size) * ori_theta1 + step_size * new_theta1
        updated_theta2 = (1.0 - step_size) * ori_theta2 + step_size * new_theta2

        # Reconstruct `mu` and `sigma` from the expectation parameters (moment matching)
        updated_mu = updated_theta1
        updated_sigma2 = updated_theta2 - updated_mu * updated_mu
        updated_sigma = tl.where(updated_sigma2 < min_sigma * min_sigma, min_sigma, tl.sqrt(updated_sigma2))

        tl.store(params_ptr + s_pids, updated_mu, mask = mask)
        tl.store(params_ptr + s_pids + 1, updated_sigma, mask = mask)

    def _get_constructor(self):
        return Gaussian, {"mu": self.mu, "sigma": self.sigma, "min_sigma": self.min_sigma}



from __future__ import annotations

import torch
import torch.nn as nn
import triton
import triton.language as tl

from pyjuice.utils.kernel_launcher import triton_jit


def compile_cum_par_flows_fn(node2tiednodes, MAX_NBLOCKS = 2048, BLOCK_SIZE = 2048):

    nblock2kernel_specs = dict()
    for source_ns, item in node2tiednodes.items():
        if len(item[0]) > 1: # If the length is 1, then everything is already accumulated in the source node's parflow
            num_par_flows = source_ns._param_flow_range[1] - source_ns._param_flow_range[0]
            pfid_start = source_ns._param_flow_range[0]
            ch_nodes = item[0]

            assert len(ch_nodes) <= MAX_NBLOCKS, f"We only support fusing at most {MAX_NBLOCKS} blocks for parameter flow accumulation. " \
                                                  "Consider setting a greater `max_tied_ns_per_parflow_block` when compiling sum layers."

            nblock = triton.next_power_of_2(len(ch_nodes))

            ch_pfids = []
            for ch_ns in ch_nodes:
                ch_pfids.append(ch_ns._param_flow_range[0])

            if nblock not in nblock2kernel_specs:
                nblock2kernel_specs[nblock] = []

            nblock2kernel_specs[nblock].append([pfid_start, num_par_flows, ch_pfids])

    kernels_args = []
    for nblock, kernel_specs in nblock2kernel_specs.items():

        BLOCK_G = nblock
        BLOCK_M = BLOCK_SIZE // BLOCK_G

        target_pfids = []
        block_sizes = []
        child_pfids = []
        for kernel_spec in kernel_specs:
            pfid_start, num_par_flows, ch_pfids = kernel_spec
            for blk_start in range(0, num_par_flows, BLOCK_M):
                blk_end = min(blk_start + BLOCK_M, num_par_flows)
                blk_size = blk_end - blk_start

                ch_pfid = [chid_start + blk_start for chid_start in ch_pfids]
                ch_pfid.extend([0] * (BLOCK_G - len(ch_pfid)))

                target_pfids.append(pfid_start + blk_start)
                block_sizes.append(blk_size)
                child_pfids.append(ch_pfid)

        target_pfids = torch.tensor(target_pfids).contiguous()
        block_sizes = torch.tensor(block_sizes).contiguous()
        child_pfids = torch.tensor(child_pfids).contiguous()

        kernels_args.append([target_pfids, block_sizes, child_pfids, BLOCK_G, BLOCK_M])

    return kernels_args


def cum_par_flows_to_device(kernels_args, device):
    for i in range(len(kernels_args)):
        target_pfids, block_sizes, ch_pfids, BLOCK_G, BLOCK_M = kernels_args[i]

        kernels_args[i] = [
            target_pfids.to(device),
            block_sizes.to(device),
            ch_pfids.to(device),
            BLOCK_G,
            BLOCK_M
        ]

    return kernels_args


@triton_jit
def cum_par_flows_kernel(param_flows, target_pfids, block_sizes, ch_pfids, BLOCK_G: tl.constexpr, BLOCK_M: tl.constexpr):

    pid = tl.program_id(axis = 0)

    offs_g = tl.arange(0, BLOCK_G) + pid * BLOCK_G
    offs_chblk = tl.load(ch_pfids + offs_g)
    mask_chblk = offs_chblk >= 0

    block_size = tl.load(block_sizes + pid)
    offs_m = tl.arange(0, BLOCK_M)
    mask_m = offs_m < block_size

    offs_chs = offs_chblk[:,None] + tl.arange(0, BLOCK_M)[None,:]
    ch_pflows = tl.load(param_flows + offs_chs, mask = mask_chblk[:,None] & mask_m[None,:], other = 0)
    
    tar_pflows = tl.sum(ch_pflows, axis = 0)

    tar_pfid = tl.load(target_pfids + pid)
    tl.store(param_flows + tar_pfid + offs_m, tar_pflows, mask = mask_m)


def compute_cum_par_flows(param_flows, kernels_args):

    for kernel_args in kernels_args:

        target_pfids, block_sizes, ch_pfids, BLOCK_G, BLOCK_M = kernel_args

        grid = (target_pfids.size(0),)

        cum_par_flows_kernel[grid](param_flows, target_pfids, block_sizes, ch_pfids, BLOCK_G, BLOCK_M)

    return None
            


from __future__ import annotations

import torch

from typing import Optional, Any


class Distribution():
    def __init__(self):
        self.post_fw_fns = []
        self.post_bp_fns = []

    def get_signature(self):
        """
        Get the signature of the current distribution.
        """
        raise NotImplementedError()

    def get_metadata(self):
        """
        Get the metadata of the current distribution.
        """
        return [] # no metadata

    def normalize_parameters(self, params: torch.Tensor, **kwargs):
        """
        Normalize node parameters.
        """
        return params

    def set_meta_parameters(self, **kwargs):
        """
        Assign meta-parameters to `self._params`.
        
        :note: the actual parameters are not initialized after this function call.
        """
        raise NotImplementedError()

    def num_parameters(self):
        """
        The number of parameters per node.
        """
        raise NotImplementedError()

    def num_param_flows(self):
        """
        The number of parameter flows per node.
        """
        raise NotImplementedError()

    def init_parameters(self, num_nodes: int, perturbation: float = 2.0, params: Optional[Any] = None, **kwargs):
        """
        Randomly initialize node parameters.

        :param num_nodes: number of nodes
        :type num_nodes: int

        :param perturbation: "amount of perturbation" added to the parameters (should be greater than 0)
        :type perturbation: float
        """
        raise NotImplementedError()

    def init_meta_parameters(self, num_nodes: int, params: Any, **kwargs):
        """
        Initialize meta-parameters for `num_nodes` nodes.
        Return shape should be the same with `init_parameters`.
        """
        raise NotImplementedError()

    @property
    def need_meta_parameters(self):
        """
        A flag indicating whether users need to pass in meta-parameters to the 
        constructor of InputNodes.
        """
        return False

    def get_data_dtype(self):
        """
        Get the data dtype for the distribution.
        """
        return torch.float32

    def get_fw_mar_fn(self):
        """
        Return the `fw_mar_fn` to be used by the layer.
        """
        return self.fw_mar_fn

    def get_bk_flow_fn(self):
        """
        Return the `bk_flow_fn` to be used by the layer.
        """
        return self.bk_flow_fn

    def get_sample_fn(self):
        """
        Return the `sample_fn` to be used by the layer.
        """
        return self.sample_fn

    def get_em_fn(self):
        """
        Return the `em_fn` to be used by the layer.
        """
        return self.em_fn

    @staticmethod
    def fw_mar_fn(*args, **kwargs):
        """
        Forward evaluation for log-probabilities.

        :param local_offsets: [BLOCK_SIZE] the local indices of the to-be-processed input nodes
        :param data: [BLOCK_SIZE, num_vars_per_node] data of the corresponding nodes
        :param params_ptr: pointer to the parameter vector
        :param s_pids: [BLOCK_SIZE] start parameter index (offset) for all input nodes
        :param metadata_ptr: pointer to metadata
        :param s_mids_ptr: pointer to the start metadata index (offset)
        :param mask: [BLOCK_SIZE] indicate whether each node should be processed
        :param num_vars_per_node: numbers of variables per input node/distribution
        :param BLOCK_SIZE: CUDA block size
        """
        raise NotImplementedError()

    @staticmethod
    def bk_flow_fn(*args, **kwargs):
        """
        Accumulate statistics and compute input parameter flows.

        :param local_offsets: [BLOCK_SIZE] the local indices of the to-be-processed input nodes
        :param ns_offsets: [BLOCK_SIZE] the global offsets used to load from `node_mars_ptr`
        :param data: [BLOCK_SIZE, num_vars_per_node] data of the corresponding nodes
        :param flows: [BLOCK_SIZE] node flows
        :param node_mars_ptr: pointer to the forward values
        :param params_ptr: pointer to the parameter vector
        :param param_flows_ptr: pointer to the parameter flow vector
        :param s_pids: [BLOCK_SIZE] start parameter index (offset) for all input nodes
        :param s_pfids: [BLOCK_SIZE] start parameter flow index (offset) for all input nodes
        :param metadata_ptr: pointer to metadata
        :param s_mids_ptr: pointer to the start metadata index (offset)
        :param mask: [BLOCK_SIZE] indicate whether each node should be processed
        :param num_vars_per_node: numbers of variables per input node/distribution
        :param BLOCK_SIZE: CUDA block size
        """
        raise NotImplementedError()

    @staticmethod
    def sample_fn(*args, **kwargs):
        """
        Sample from the distribution.

        :param samples_ptr: pointer to store the resultant samples 
        :param local_offsets: [BLOCK_SIZE] the local indices of the to-be-processed input nodes
        :param batch_offsets: [BLOCK_SIZE] batch id corresponding to every node
        :param vids: [BLOCK_SIZE] variable ids (only univariate distributions are supported)
        :param s_pids: [BLOCK_SIZE] start parameter index (offset) for all input nodes 
        :param params_ptr: pointer to the parameter vector
        :param metadata_ptr: pointer to metadata
        :param s_mids_ptr: pointer to the start metadata index (offset)
        :param mask: [BLOCK_SIZE] indicate whether each node should be processed
        :param batch_size: batch size
        :param BLOCK_SIZE: CUDA block size
        :param seed: random seed
        """
        raise NotImplementedError()

    @staticmethod
    def em_fn(*args, **kwargs):
        """
        Parameter update with EM

        :param local_offsets: [BLOCK_SIZE] the local indices of the to-be-processed input nodes
        :param params_ptr: pointer to the parameter vector
        :param param_flows_ptr: pointer to the parameter flow vector
        :param s_pids: [BLOCK_SIZE] start parameter index (offset) for all input nodes
        :param s_pfids: [BLOCK_SIZE] start parameter flow index (offset) for all input nodes
        :param metadata_ptr: pointer to metadata
        :param s_mids_ptr: pointer to the start metadata index (offset)
        :param mask: [BLOCK_SIZE] indicate whether each node should be processed
        :param step_size: EM step size (0, 1]
        :param pseudocount: pseudocount 
        :param BLOCK_SIZE: CUDA block size
        """
        raise NotImplementedError()

    @staticmethod
    def partition_fn(local_offsets, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, BLOCK_SIZE, TILE_SIZE_K):
        """
        Evaluate partition function

        :param local_offsets: [BLOCK_SIZE] the local indices of the to-be-processed input nodes
        :param params_ptr: pointer to the parameter vector
        :param s_pids: [BLOCK_SIZE] start parameter index (offset) for all input nodes
        :param metadata_ptr: pointer to metadata
        :param s_mids_ptr: pointer to the start metadata index (offset)
        :param mask: [BLOCK_SIZE] indicate whether each node should be processed
        :param BLOCK_SIZE: CUDA block size
        :param TILE_SIZE_K: tile size for processing each individual input node
        """
        raise NotImplementedError()

    def _get_constructor(self):
        raise NotImplementedError()

    def _need_2nd_kernel_dim(self):
        return False



import torch
import torch.nn as nn
from torch.optim.optimizer import Optimizer
from typing import Dict, Optional

from pyjuice.model import TensorCircuit


class CircuitOptimizer():

    SUPPORTED_OPTIM_METHODS = ["EM", "Viterbi", "GeneralEM"]

    def __init__(self, pc: TensorCircuit, base_optimizer: Optional[Optimizer] = None, method: str = "EM", lr: float = 0.1,
                 pseudocount: float = 0.1, **kwargs):
        
        self.pc = pc

        self.base_optimizer = base_optimizer

        assert method in self.SUPPORTED_OPTIM_METHODS, f"Unsupported optimization method {method} for PCs."
        self.method = method

        self.lr = lr
        self.pseudocount = pseudocount

    def zero_grad(self):
        if self.base_optimizer is not None:
            self.base_optimizer.zero_grad()

        self.pc.init_param_flows(flows_memory = 0.0)

    def step(self, closure = None):
        if self.base_optimizer is not None:
            self.base_optimizer.step()

        if self.method == "EM":
            self.pc.mini_batch_em(
                step_size = self.lr,
                pseudocount = self.pseudocount
            )
        else:
            raise ValueError(f"Unknown PC optimization method {self.method}.")

    def state_dict(self):
        if self.base_optimizer is not None:
            state_dict = self.base_optimizer.state_dict()
        else:
            state_dict = dict()

        state_dict["pc_states"] = {
            "method": self.method, 
            "lr": self.lr, 
            "pseudocount": self.pseudocount
        }

    def load_state_dict(self, state_dict: Dict):
        pc_states = state_dict["pc_states"]
        
        self.method = pc_states["method"]
        self.lr = pc_states["lr"]
        self.pseudocount = pc_states["pseudocount"]

        del state_dict["pc_states"]

        if self.base_optimizer is not None:
            self.base_optimizer.load_state_dict(state_dict)

    @property
    def param_groups(self):
        assert self.base_optimizer is not None
        return self.base_optimizer.param_groups